#include "Database.h"
#include <stdio.h>

HANDLE g_hStopDBSLoopEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

DatabaseServer::DatabaseServer()
{
	// Start the Winsock dll
	WSADATA wsaData;
	int err = WSAStartup( MAKEWORD( 2, 0 ), &wsaData );
	if (err != 0)
		printf("Winsock2 dll not initialized\n");

	m_hServerThread = NULL;
	m_nPort = 0;
	strcpy(m_pszHost, "127.0.0.1");
	gethostname(m_pszHost, 100);
	m_pList = NULL;
}

void DeleteValueList(DatabaseServer::ValueNode *pList)
{
	if (pList == NULL)
		return;
	DeleteValueList(pList->pNext);
	if (pList->pData != NULL)
		delete pList->pData;
	delete pList;
}
void DeleteKeyList(DatabaseServer::KeyNode *pList)
{
	if (pList == NULL)
		return;
	DeleteKeyList(pList->pNext);
	DeleteValueList(pList->pValueList);
	if (pList->pszKey != NULL)
		delete pList->pszKey;
	delete pList;
}
void DeleteIDList(DatabaseServer::IDNode *pList)
{
	if (pList == NULL)
		return;
	DeleteIDList(pList->pNext);
	DeleteKeyList(pList->pKeyList);
	delete pList;
}

DatabaseServer::~DatabaseServer()
{
	if (m_hServerThread != NULL)
	{
		//TerminateThread(m_hServerThread, 0);
		//m_hServerThread = NULL;
		SetEvent(g_hStopDBSLoopEvent);
		while (m_hServerThread)
			Sleep(200);
	}
	CloseHandle(g_hStopDBSLoopEvent);
	g_hStopDBSLoopEvent = NULL;
	if (m_pList != NULL)
		DeleteIDList(m_pList);
	m_pList = NULL;
}

bool DatabaseServer::Start()
{
	if (m_hServerThread == NULL)
	{
		DWORD dwThreadID;
		m_hServerThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)DatabaseServerThread, this, 0, &dwThreadID);
		if (m_hServerThread == NULL)
		{
			return false;
		}
	}
	return true;
}

bool DatabaseServer::GetHost(char *pszHost, int length)
{
	if (strlen(m_pszHost) >= (unsigned int)length)
		return false;
	strcpy(pszHost, m_pszHost);
	return true;
}

bool DatabaseServer::SetPort(int nPort)
{
	if (m_hServerThread != NULL)
		return false;
	m_nPort = nPort;
	return true;
}

int DatabaseServer::GetPort()
{
	if (m_hServerThread != NULL)
	{
		while (m_nPort == 0)
			Sleep(200);
		return m_nPort;
	}
	return -1;
}

bool DatabaseServer::Stop()
{
	if (m_hServerThread != NULL)
	{
		//TerminateThread(m_hServerThread, 0);
		//m_hServerThread = NULL;
		SetEvent(g_hStopDBSLoopEvent);
		while (m_hServerThread)
			Sleep(200);
	}
	return true;
}

int DatabaseServer::Delete(char *pszID)
{
	IDNode **ppNode;

	ppNode = &m_pList;
	while (*ppNode != NULL)
	{
		if (strcmp((*ppNode)->pszID, pszID) == 0)
		{

			DeleteKeyList( (KeyNode*)InterlockedExchange( (long*)&(*ppNode)->pKeyList, NULL ) );
			return MPI_DBS_SUCCESS;
		}
		ppNode = &(*ppNode)->pNext;
	}

	return MPI_DBS_SUCCESS;
}

/*
int DatabaseServer::Delete(char *pszID)
{
	IDNode **ppNode, *pNode;
	KeyNode *ppKey;

	ppNode = &m_pList;
	while (*ppNode != NULL)
	{
		pNode = (IDNode *)InterlockedExchange( (long*)ppNode, NULL );
		if (strcmp(pNode->pszID, pszID) == 0)
		{

			DeleteKeyList( (KeyNode*)InterlockedExchange( (long*)&(*ppNode)->pKeyList, NULL ) );
			return MPI_DBS_SUCCESS;
		}
		else
		{

			InterlockedExchange( (long*)&(*ppNode)->pNext, pNode );
		}
		ppNode = &(*ppNode)->pNext;
	}

	return MPI_DBS_SUCCESS;
}
//*/

int DatabaseServer::Get(char *pszID, char *pszKey, void *&pValueData, int *length)
{
	IDNode **ppNode;
	KeyNode **ppKey;
	ValueNode *pValue;

	ppNode = &m_pList;
	while (true)
	{
		while (*ppNode == NULL)
			Sleep(100);
		if (strcmp((*ppNode)->pszID, pszID) == 0)
		{
			ppKey = &(*ppNode)->pKeyList;
			while (true)
			{
				while (*ppKey == NULL)
					Sleep(100);
				if (strcmp((*ppKey)->pszKey, pszKey) == 0)
				{
					while ( (pValue = (ValueNode*)InterlockedExchange( (long*)(&(*ppKey)->pValueList), NULL)) == NULL)
						Sleep(100);
					if (pValue->length > *length)
					{
						*length = pValue->length;
						InterlockedExchange( (long*)&(*ppKey)->pValueList, (long)pValue);
						return MPI_DBS_FAIL;
					}
					if ((*ppKey)->bPersistent)
					{
						// make a copy of the data
						pValueData = new char[pValue->length];
						memcpy(pValueData, pValue->pData, pValue->length);
						*length = pValue->length;
						// put the data back
						(*ppKey)->pValueList = pValue;
					}
					else
					{
						// consume the data
						pValueData = pValue->pData;
						*length = pValue->length;
						// put the next entry as the new head of the list
						(*ppKey)->pValueList = pValue->pNext;
					}
					return MPI_DBS_SUCCESS;
				}
				ppKey = &(*ppKey)->pNext;
			}
		}
		ppNode = &(*ppNode)->pNext;
	}

	return MPI_DBS_SUCCESS;
}

int DatabaseServer::Put(char *pszID, char *pszKey, void *pValueData, int length, bool bPersistent)
{
	PVOID pNull = NULL;
	IDNode *pNode = new IDNode, **ppIter = &m_pList;
	pNode->pKeyList = NULL;
	pNode->pNext = NULL;
	pNode->pPrev = NULL;
	strcpy(pNode->pszID, pszID);

	ppIter = &m_pList;
	while (InterlockedCompareExchange((void**)ppIter, pNode, pNull) != NULL)
	{
		if (strcmp((*ppIter)->pszID, pszID) == 0)
		{
			delete pNode;
			KeyNode *pKey = new KeyNode;
			pKey->pNext = NULL;
			pKey->bPersistent = bPersistent;
			pKey->pszKey = pszKey;
			pKey->pValueList = NULL;
			KeyNode **ppKey = &(*ppIter)->pKeyList;
			while (InterlockedCompareExchange( (void**)ppKey, pKey, pNull) != NULL)
			{
				if (strcmp((*ppKey)->pszKey, pszKey) == 0)
				{
					delete pszKey;
					delete pKey;
					ValueNode *pValue = new ValueNode;
					pValue->length = length;
					pValue->pData = pValueData;
					pValue->pNext = NULL;
					//if ((*ppKey)->bPersistent) // <----- This method retains the option specified on the first put call
					(*ppKey)->bPersistent = bPersistent; // <---- This method sets the state to match the current call
					if (bPersistent)
					{
						DeleteValueList( (ValueNode*)InterlockedExchange( (long*)&(*ppKey)->pValueList, (long)pValue));
					}
					else
					{
						ValueNode **ppValue = &(*ppKey)->pValueList;
						while (InterlockedCompareExchange( (void**)ppValue, pValue, pNull) != NULL)
							ppValue = &(*ppValue)->pNext;
					}
					return MPI_DBS_SUCCESS;
				}
				ppKey = &(*ppKey)->pNext;
			}
			ValueNode *pValue = new ValueNode;
			pValue->length = length;
			pValue->pData = pValueData;
			pValue->pNext = NULL;
			(*ppKey)->pValueList = pValue;
			return MPI_DBS_SUCCESS;
		}
		pNode->pPrev = *ppIter;
		ppIter = &(*ppIter)->pNext;
	}
	KeyNode *pKey = new KeyNode;
	ValueNode *pValue = new ValueNode;
	pValue->pNext = NULL;
	pValue->pData = pValueData;
	pValue->length = length;
	pKey->bPersistent = bPersistent;
	pKey->pNext = NULL;
	pKey->pValueList = pValue;
	pKey->pszKey = pszKey;
	pNode->pKeyList = pKey;

	return MPI_DBS_SUCCESS;
}
/*
int DatabaseServer::Put(char *pszID, char *pszKey, void *pValueData, int length, bool bPersistent)
{
	PVOID pNull = NULL;
	IDNode *pNode = new IDNode, **ppIter = &m_pList;
	pNode->pKeyList = NULL;
	pNode->pNext = NULL;
	pNode->pPrev = NULL;
	strcpy(pNode->pszID, pszID);

	ppIter = &m_pList;
	while (InterlockedCompareExchange((void**)&ppIter, pNode, pNull) != NULL)
	{
		if (strcmp((*ppIter)->pszID, pszID) == 0)
		{
			delete pNode;
			KeyNode *pKey = new KeyNode;
			pKey->pNext = NULL;
			pKey->bPersistent = bPersistent;
			pKey->pszKey = pszKey;
			pKey->pValueList = NULL;
			KeyNode **ppKey = &(*ppIter)->pKeyList;
			while (InterlockedCompareExchange( (void**)&ppKey, &pKey, &pNull) != NULL)
			{
				if (strcmp((*ppKey)->pszKey, pszKey) == 0)
				{
					delete pszKey;
					delete pKey;
					ValueNode *pValue = new ValueNode;
					pValue->length = length;
					pValue->pData = pValueData;
					pValue->pNext = NULL;
					if ((*ppKey)->bPersistent)
					{
						DeleteValueList( (ValueNode*)InterlockedExchange( (long*)&(*ppKey)->pValueList, (long)pValue));
					}
					else
					{
						ValueNode **ppValue = &(*ppKey)->pValueList;
						while (InterlockedCompareExchange( (void**)&ppValue, &pValue, &pNull) != NULL)
							ppValue = &(*ppValue)->pNext;
					}
					return MPI_DBS_SUCCESS;
				}
				ppKey = &(*ppKey)->pNext;
			}
			ValueNode *pValue = new ValueNode;
			pValue->length = length;
			pValue->pData = pValueData;
			pValue->pNext = NULL;
			(*ppKey)->pValueList = pValue;
			return MPI_DBS_SUCCESS;
		}
		pNode->pPrev = *ppIter;
		ppIter = &(*ppIter)->pNext;
	}
	KeyNode *pKey = new KeyNode;
	ValueNode *pValue = new ValueNode;
	pValue->pNext = NULL;
	pValue->pData = pValueData;
	pValue->length = length;
	pKey->bPersistent = bPersistent;
	pKey->pNext = NULL;
	pKey->pValueList = pValue;
	pNode->pKeyList = pKey;

	return MPI_DBS_SUCCESS;
}
//*/

void DatabaseServer::PrintState()
{
	IDNode *pNode = m_pList;
	KeyNode *pKey;
	ValueNode *pValue;

	printf("DATABASE:\n{\n");
	while (pNode)
	{
		printf(" NODE \"%s\"\n", pNode->pszID);
		pKey = pNode->pKeyList;
		while (pKey)
		{
			if (pKey->bPersistent)
				printf("  KEY \"%s\" persistent\n", pKey->pszKey);
			else
				printf("  KEY \"%s\" consumable\n", pKey->pszKey);
			pValue = pKey->pValueList;
			while (pValue)
			{
				printf("   VALUE: ");
				fwrite(pValue->pData, pValue->length, 1, stdout);
				printf("\n");
				pValue = pValue->pNext;
			}
			pKey = pKey->pNext;
		}
		pNode = pNode->pNext;
	}
	printf("}\n");
}
