	
// MPILaunch.h : Declaration of the CMPILaunch

#ifndef __MPILAUNCH_H_
#define __MPILAUNCH_H_

#include "resource.h"       // main symbols
#include "ChunkNode.h"
#include "..\Common\MPIJobLog.h"

extern HANDLE g_hLaunchSyncMutex;

/////////////////////////////////////////////////////////////////////////////
// CMPILaunch
class ATL_NO_VTABLE CMPILaunch : 
	public CComObjectRootEx<CComMultiThreadModel>,
	public CComCoClass<CMPILaunch, &CLSID_MPILaunch>,
	public IDispatchImpl<IMPILaunch, &IID_IMPILaunch, &LIBID_MPILAUNCHSERVERLib>
{
public:
	CMPILaunch()
	{
		m_pUnkMarshaler = NULL;
		m_hProcess = NULL;
		m_hOutputMutex = CreateMutex(NULL, FALSE, NULL);
		m_hOutputEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
		m_hInteractivePipeR = NULL;
		m_hInteractivePipeW = NULL;
		m_pOutList = NULL;
		m_pOutListTail = NULL;
		m_hOutputThread = NULL;
		m_dwProcessId = 0;
		m_bLaunchOnDesktop = false;
	}
	~CMPILaunch();

DECLARE_REGISTRY_RESOURCEID(IDR_MPILAUNCH)
DECLARE_GET_CONTROLLING_UNKNOWN()

DECLARE_PROTECT_FINAL_CONSTRUCT()

BEGIN_COM_MAP(CMPILaunch)
	COM_INTERFACE_ENTRY(IMPILaunch)
	COM_INTERFACE_ENTRY(IDispatch)
	COM_INTERFACE_ENTRY_AGGREGATE(IID_IMarshal, m_pUnkMarshaler.p)
END_COM_MAP()

	HRESULT FinalConstruct()
	{
		return CoCreateFreeThreadedMarshaler(
			GetControllingUnknown(), &m_pUnkMarshaler.p);
	}

	void FinalRelease()
	{
		m_pUnkMarshaler.Release();
	}

	CComPtr<IUnknown> m_pUnkMarshaler;

// IMPILaunch
public:
	STDMETHOD(GrantAccessToDesktop)(BSTR bAccount, BSTR bPassword, long *error, BSTR *berror_msg);
	STDMETHOD(CreateTempFile)(BSTR *bFilename, long *error, BSTR *berror_msg);
	STDMETHOD(GetPortFromFile)(BSTR bFilename, long *nPort, long *error, BSTR *berror_msg);
	STDMETHOD(LaunchInteractiveImpersontatedMPIProcess)(BSTR bCmdLine, BSTR bDir, BSTR bEnv, long *pid, long *error, BSTR *berror_msg);
	STDMETHOD(SendBreak)(long *error, BSTR *berror_msg);
	STDMETHOD(PutInteractiveInput)(VARIANT vInput, long *error, BSTR *berror_msg);
	STDMETHOD(Abort)(long *error, BSTR *berror_msg);
	STDMETHOD(LaunchMPIProcess)(BSTR bCmdLine, BSTR bDir, BSTR bEnv, BSTR bAccount, BSTR bPassword, long *pid, long *error, BSTR *berror_msg);
	STDMETHOD(GetInteractiveOutput)(VARIANT *vOutput, long *more, long *error, BSTR *berror_msg);
	STDMETHOD(LaunchInteractiveMPIProcess)(BSTR bCmdLine, BSTR bDir, BSTR bEnv, BSTR bAccount, BSTR bPassword, long *pid, long *error, BSTR *berror_msg);

	HANDLE m_hProcess, m_hOutputMutex, m_hInteractivePipeR, m_hInteractivePipeW;
	DWORD m_dwProcessId;
	HANDLE m_hOutputThread, m_hOutputEvent;
	ChunkNode *m_pOutList, *m_pOutListTail;
	bool m_bLaunchOnDesktop;
};

#endif //__MPILAUNCH_H_
