#include <stdio.h>
#include <tchar.h>
#include "..\Common\MPICH_pwd.h"
#include "..\Common\MPIJobDefs.h"
#include <conio.h>
// MPIJobServer must be compiled before the following files are generated
#include "..\MPILaunchServer\MPILaunchServer.h"
#include <objbase.h>
#include "..\MPILaunchServer\MPILaunchserver_i.c"
#include <time.h>
#include "..\Common\Translate_Error.h"

// Global variables
long g_nNproc = 1;
VARIANT g_vHosts, g_vSMPInfo;
TCHAR g_pszExe[MAX_PATH] = _T(""), g_pszArgs[MAX_PATH] = _T("");
TCHAR g_pszFirstHost[100] = _T("");
IMPILaunch *g_pLaunch = NULL;
HANDLE g_hFinishedEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

// Function name	: PrintError
// Description	    : 
// Return type		: void 
// Argument         : HRESULT hr
void PrintError(HRESULT hr)
{
	HLOCAL str;
	FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ALLOCATE_BUFFER,
		0,
		hr,
		MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ),
		(LPTSTR) &str,
		0,0);
	
	_tprintf(TEXT("error: %s\n"), str);
	LocalFree(str);
}

bool Connect(LPTSTR host, IMPILaunch *&pLaunch)
{
	HRESULT hr;
	MULTI_QI qi;
	WCHAR wHost[100], localhost[100];
	COSERVERINFO server;
	DWORD length = 100;

	hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);

	if (FAILED(hr))
	{
		_tprintf(TEXT("CoInitialize() failed.\n"));
		PrintError(hr);
		return false;
	}

	//*
	hr = CoInitializeSecurity(NULL, -1, NULL, NULL,
		RPC_C_AUTHN_LEVEL_CONNECT, RPC_C_IMP_LEVEL_IMPERSONATE, NULL, EOAC_NONE, NULL);
		//RPC_C_AUTHN_LEVEL_PKT, RPC_C_IMP_LEVEL_IMPERSONATE, NULL, EOAC_NONE, NULL);
	if (FAILED(hr))
	{
		if (hr == RPC_E_TOO_LATE)
			printf("CoInitializeSecurity failed in Connect(MPIJob) because it has already been set.\n");
		else
		{
			char error_msg[256];
			Translate_HRError(hr, error_msg);
			printf("CoInitializeSecurity failed in Connect(MPIJob)\nError: %s", error_msg);
		}
	}
	//*/

	qi.pIID = &IID_IMPILaunch;
	qi.pItf = NULL;

#ifdef UNICODE
	wcscpy(wHost, host);
#else
	//MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, host, -1, wHost, 100);
	mbstowcs(wHost, host, strlen(host)+1);
#endif

	server.dwReserved1 = 0;
	server.dwReserved2 = 0;
	server.pAuthInfo = NULL;
	server.pwszName = wHost;

	GetComputerNameW(localhost, &length);

	if (_wcsicmp(localhost, wHost) == 0)
	{
		hr = CoCreateInstanceEx(CLSID_MPILaunch, NULL, CLSCTX_SERVER, &server, 1, &qi);
		if (FAILED(hr))
		{
			_tprintf(TEXT("Unable to connect to %s: "), host);
			PrintError(hr);
			return false;
		}
	}
	else
	{
		hr = CoCreateInstanceEx(CLSID_MPILaunch, NULL, CLSCTX_REMOTE_SERVER, &server, 1, &qi);
		if (FAILED(hr))
		{
			_tprintf(TEXT("Unable to connect to %s: "), host);
			PrintError(hr);
			return false;
		}
	}

	if (SUCCEEDED(qi.hr))
	{
		pLaunch = (IMPILaunch*)qi.pItf;
		return true;
	}

	return false;
}

// Function name	: PrintOptions
// Description	    : 
// Return type		: void 
void PrintOptions()
{
	_tprintf(TEXT("Usage:\n"));
	_tprintf(TEXT("   msh host\n"));
}

// Function name	: CtrlHandlerRoutine
// Description	    : 
// Return type		: BOOL WINAPI 
// Argument         : DWORD dwCtrlType
//bool g_bFirst = true;
BOOL WINAPI CtrlHandlerRoutine(DWORD dwCtrlType)
{
	fprintf(stderr, "User break\n");
	if (g_pLaunch != NULL)
	{
		BSTR berror_msg = SysAllocString(L"");
		LONG error;
		HRESULT hr;
		/*
		if (g_bFirst)
		{
			fprintf(stderr, "Forwarding break signal\n");
			g_pLaunch->SendBreak(&error, &berror_msg);
			SysFreeString(berror_msg);
			g_bFirst = false;
		}
		else
		//*/
		{
			error = 0;
			hr = g_pLaunch->Abort(&error, &berror_msg);
			if (FAILED(hr))
			{
				printf("Abort failed.\n");
				PrintError(hr);
			}
			if (error)
				wprintf(L"Abort failed: %s", berror_msg);
			SysFreeString(berror_msg);
			g_pLaunch->Release();
			ExitProcess(1);
		}
	}
	return TRUE;
}

struct RedirectInputThreadArg
{
	HANDLE hEvent;
	IStream **ppStream;
};

char g_pBuffer[1024];
HANDLE g_hBufferEvent1 = CreateEvent(NULL, TRUE, FALSE, NULL);
HANDLE g_hBufferEvent2 = CreateEvent(NULL, TRUE, FALSE, NULL);
DWORD g_num_read = 0;

// Function name	: ReadStdinThread
// Description	    : 
// Return type		: void 
void ReadStdinThread()
{
	HANDLE hStdin = GetStdHandle(STD_INPUT_HANDLE);
	while (true)
	{
		if (!ReadFile(hStdin, g_pBuffer, 1024, &g_num_read, NULL))
			return;
		if (g_num_read == 0)
			return;
		ResetEvent(g_hBufferEvent2);
		SetEvent(g_hBufferEvent1);
		WaitForSingleObject(g_hBufferEvent2, INFINITE);
	}
}

// Function name	: RedirectInputThread
// Description	    : 
// Return type		: void 
// Argument         : RedirectInputThreadArg *arg
void RedirectInputThread(RedirectInputThreadArg *arg)
{
	IMPILaunch *pLaunch=NULL;
	HRESULT hr=S_OK;
	HANDLE hObject[2];
	long error=0;
	BSTR berror_msg;
	DWORD ret_val;

	hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);

	berror_msg = SysAllocString(L"");
	CoGetInterfaceAndReleaseStream (*arg->ppStream, IID_IMPILaunch, (void**) &pLaunch);
	delete arg->ppStream;

	DWORD dwThreadID;
	HANDLE hRSIThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)ReadStdinThread, NULL, 0, &dwThreadID);

	hObject[0] = arg->hEvent;
	hObject[1] = g_hBufferEvent1;

	berror_msg = SysAllocString(L"");
	while (true)
	{
		ret_val = WaitForMultipleObjects(2, hObject, FALSE, INFINITE);
		if (ret_val == WAIT_OBJECT_0+1)
		{
			if (g_num_read > 0)
			{
				SAFEARRAYBOUND bound;
				VARIANT vInput;
				void *pBuf;

				VariantInit(&vInput);
				bound.lLbound = 0;
				bound.cElements = g_num_read;
				vInput.vt = VT_UI1 | VT_ARRAY;
				vInput.parray = SafeArrayCreate(VT_UI1, 1, &bound);

				SafeArrayAccessData(vInput.parray, &pBuf);
				memcpy(pBuf, g_pBuffer, g_num_read);
				SafeArrayUnaccessData(vInput.parray);

				error = 0;
				hr = pLaunch->PutInteractiveInput(vInput, &error, &berror_msg);
				if (FAILED(hr))
				{
					VariantClear(&vInput);
					printf("PutInteractiveInput failed: %d\n", hr);
					PrintError(hr);
					break;
				}
				if (error)
				{
					VariantClear(&vInput);
					if (wcslen(berror_msg) < 1)
						wprintf(L"PutInteractiveInput failed: %d %s\n", error, berror_msg);
					else
						wprintf(L"PutInteractiveInput failed: %s\n", berror_msg);
					break;
				}
				VariantClear(&vInput);
			}
			ResetEvent(g_hBufferEvent1);
			SetEvent(g_hBufferEvent2);
		}
		else
		{
			//printf("g_hFinishedEvent signalled\n");
			break;
		}
	}

	pLaunch->Release();
	CoUninitialize();
}

// Function name	: main
// Description	    : 
// Return type		: void 
// Argument         : int argc
// Argument         : TCHAR *argv[]
void main(int argc, TCHAR *argv[])
{
	SetConsoleCtrlHandler(CtrlHandlerRoutine, TRUE);

	if (argc == 1)
	{
		PrintOptions();
		return;
	}

	_tcscpy(g_pszFirstHost, argv[1]);

	if (argc == 2)
		_tcscpy(g_pszExe, TEXT("cmd.exe /Q"));
	else
	{
		_tcscpy(g_pszExe, TEXT("cmd.exe /c "));
		for (int i=2; i<argc; i++)
		{
			_tcscat(g_pszExe, argv[i]);
			if (i < argc-1)
				_tcscat(g_pszExe, TEXT(" "));
		}
	}

	TCHAR pszDir[MAX_PATH] = TEXT(".");
	GetCurrentDirectory(MAX_PATH, pszDir);

	LONG error = 0;
	BSTR berror_msg;
	BSTR bCommonArgs, bExe, bDir, bEnv;
	HRESULT hr = S_OK;

	berror_msg = SysAllocString(L"");
	bEnv = SysAllocString(L"");
#ifdef UNICODE
	bExe = SysAllocString(g_pszExe);
	bCommonArgs = SysAllocString(g_pszArgs);
	bDir = SysAllocString(pszDir);
#else
	WCHAR wTemp[MAX_PATH];
	mbstowcs(wTemp, g_pszExe, strlen(g_pszExe)+1);
	bExe = SysAllocString(wTemp);
	mbstowcs(wTemp, g_pszArgs, strlen(g_pszArgs)+1);
	bCommonArgs = SysAllocString(wTemp);
	mbstowcs(wTemp, pszDir, strlen(pszDir)+1);
	bDir = SysAllocString(wTemp);
#endif

	if (Connect(g_pszFirstHost, g_pLaunch))
	{
		error = 0;
		long pid;
		hr = g_pLaunch->LaunchInteractiveImpersontatedMPIProcess(bExe, bDir, bEnv, &pid, &error, &berror_msg);
		
		if (SUCCEEDED(hr))
		{
			if (error)
			{
				if (wcslen(berror_msg) < 1)
					wprintf(L"LaunchInteractiveImpersontatedMPIProcess failed: Error(%d), %s", error, berror_msg);
				else
					wprintf(L"%s", berror_msg);
				g_pLaunch->Release();
				CoUninitialize();
				SysFreeString(berror_msg);
				SysFreeString(bExe);
				SysFreeString(bCommonArgs);
				SysFreeString(bDir);
				return;
			}
			
			error = 0;
			
			RedirectInputThreadArg *pArg = new RedirectInputThreadArg;
			pArg->hEvent = g_hFinishedEvent;
			pArg->ppStream = new IStream*;
			// Marshall pLaunch to a thread which redirects user input
			hr = CoMarshalInterThreadInterfaceInStream(IID_IMPILaunch, g_pLaunch, pArg->ppStream);
			if (FAILED(hr))
			{
				g_pLaunch->Abort(&error, &berror_msg);
				g_pLaunch->Release();
				SysFreeString(berror_msg);
				printf("CoMarshalInterThreadInterfaceInStream failed.\n");
				CoUninitialize();
				delete pArg;
				return;
			}
			
			DWORD dwThreadID;
			HANDLE hRedirectInputThread = CreateThread(
				NULL, 0, 
				(LPTHREAD_START_ROUTINE)RedirectInputThread,
				pArg, 
				0, &dwThreadID);
			
			if (hRedirectInputThread == NULL)
			{
				error = GetLastError();
				printf("CreateThread failed: error %d\n", error);
				g_pLaunch->Abort(&error, &berror_msg);
				g_pLaunch->Release();
				SysFreeString(berror_msg);
				CoUninitialize();
				delete pArg;
				return;
			}
			
			HANDLE hStdout;
			hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
			if (hStdout == INVALID_HANDLE_VALUE)
			{
				error = GetLastError();
				_tprintf(TEXT("GetStdHandle failed: Error %d\n"), error);
				g_pLaunch->Abort(&error, &berror_msg);
				g_pLaunch->Release();
				CoUninitialize();
				SysFreeString(berror_msg);
				SysFreeString(bExe);
				SysFreeString(bCommonArgs);
				SysFreeString(bDir);
				return;
			}
			
			long more = 1, from = 0;
			VARIANT *v = new VARIANT;
			DWORD num_written, num_elements;
			void *pBuf;
			
			v->vt = VT_UI1 | VT_ARRAY;
			SAFEARRAYBOUND bound;
			bound.lLbound = 0;
			bound.cElements = 0;
			v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
			
			while (more == 1)
			{
				error = 0;
				hr = g_pLaunch->GetInteractiveOutput(v, &more, &error, &berror_msg);
				if (FAILED(hr))
				{
					_tprintf(TEXT("DCOM failure: GetJobOutput()\n"));
					PrintError(hr);
					SysFreeString(berror_msg);
					SysFreeString(bExe);
					SysFreeString(bCommonArgs);
					SysFreeString(bDir);
					VariantClear(v);
					delete v;
					return;
				}
				if (error)
				{
					if (wcslen(berror_msg) < 1)
						wprintf(L"Error %d: %s", error, berror_msg);
					else
						wprintf(L"%s", berror_msg);
					SysFreeString(berror_msg);
					SysFreeString(bExe);
					SysFreeString(bCommonArgs);
					SysFreeString(bDir);
					VariantClear(v);
					delete v;
					return;
				}
				if (v->parray != NULL)
				{
					num_elements = v->parray->rgsabound->cElements;
					if (num_elements > 0)
					{
						SafeArrayAccessData(v->parray, &pBuf);
						WriteFile(hStdout, pBuf, num_elements, &num_written, NULL);
						FlushFileBuffers(hStdout);
						SafeArrayUnaccessData(v->parray);
						// Destroy the array.
						SafeArrayDestroy(v->parray);
						// Create a new and empty array.
						v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
					}
					else
					{
						SafeArrayDestroy(v->parray);
						v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
					}
				}
				else
				{
					v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
				}
			}
			
		SafeArrayDestroy(v->parray);
		delete v;
			
		// Signal the redirect input thread to terminate
		SetEvent(g_hFinishedEvent);
		WaitForSingleObject(hRedirectInputThread, 1000);
		CloseHandle(hRedirectInputThread);
		}
		else
		{
			_tprintf(TEXT("LaunchInteractiveImpersonatedMPIProcess failed: "));
			PrintError(hr);
			fflush(stdout);
		}
			
		//g_pLaunch->Abort(&error, &berror_msg);
		g_pLaunch->Release();
		
		CoUninitialize();
	}
	else
	{
		_tprintf(TEXT("Unable to connect to '%s'\n"), g_pszFirstHost);
	}
	
	SysFreeString(berror_msg);
	SysFreeString(bExe);
	SysFreeString(bCommonArgs);
	SysFreeString(bDir);
}
