#include <stdio.h>
#include <tchar.h>
#include "..\Common\MPICH_pwd.h"
#include "..\Common\MPIJobDefs.h"
#include <conio.h>
// MPIJobServer must be compiled before the following files are generated
#include "..\MPIJobServer\MPIJobServer.h"
#include <objbase.h>
#include "..\MPIJobServer\MPIJobserver_i.c"
#include <time.h>
#include "..\Common\Translate_Error.h"

// Global variables
long g_nNproc = 1;
VARIANT g_vHosts, g_vSMPInfo;
TCHAR g_pszAccount[100], g_pszPassword[100];
TCHAR g_pszExe[MAX_PATH] = _T(""), g_pszArgs[MAX_PATH] = _T("");
TCHAR g_pszFirstHost[100] = _T("");
IMPIJob *g_pJob = NULL;
HANDLE g_hFinishedEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

// Function name	: PrintError
// Description	    : 
// Return type		: void 
// Argument         : HRESULT hr
void PrintError(HRESULT hr)
{
	HLOCAL str;
	FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ALLOCATE_BUFFER,
		0,
		hr,
		MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ),
		(LPTSTR) &str,
		0,0);
	
	_tprintf(TEXT("error: %s\n"), str);
	LocalFree(str);
}

// Function name	: Connect
// Description	    : 
// Return type		: bool 
// Argument         : LPTSTR host
// Argument         : IMPIJob *&pJob
bool Connect(LPTSTR host, IMPIJob *&pJob)
{
	HRESULT hr;
	MULTI_QI qi;
	WCHAR wHost[100], localhost[100];
	COSERVERINFO server;
	DWORD length = 100;

	hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);

	if (FAILED(hr))
	{
		_tprintf(TEXT("CoInitialize() failed.\n"));
		PrintError(hr);
		return false;
	}

	//*
	hr = CoInitializeSecurity(NULL, -1, NULL, NULL,
		RPC_C_AUTHN_LEVEL_CONNECT, RPC_C_IMP_LEVEL_IMPERSONATE, NULL, EOAC_NONE, NULL);
		//RPC_C_AUTHN_LEVEL_PKT, RPC_C_IMP_LEVEL_IMPERSONATE, NULL, EOAC_NONE, NULL);
	if (FAILED(hr))
	{
		if (hr == RPC_E_TOO_LATE)
			printf("CoInitializeSecurity failed in Connect(MPIJob) because it has already been set.\n");
		else
		{
			char error_msg[256];
			Translate_HRError(hr, error_msg);
			printf("CoInitializeSecurity failed in Connect(MPIJob)\nError: %s", error_msg);
		}
	}
	//*/

	qi.pIID = &IID_IMPIJob;
	qi.pItf = NULL;

#ifdef UNICODE
	wcscpy(wHost, host);
#else
	//MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, host, -1, wHost, 100);
	mbstowcs(wHost, host, strlen(host)+1);
#endif

	server.dwReserved1 = 0;
	server.dwReserved2 = 0;
	server.pAuthInfo = NULL;
	server.pwszName = wHost;

	GetComputerNameW(localhost, &length);

	if (_wcsicmp(localhost, wHost) == 0)
	{
		hr = CoCreateInstanceEx(CLSID_MPIJob, NULL, CLSCTX_SERVER, &server, 1, &qi);
		if (FAILED(hr))
		{
			_tprintf(TEXT("Unable to connect to %s: "), host);
			PrintError(hr);
			return false;
		}
	}
	else
	{
		hr = CoCreateInstanceEx(CLSID_MPIJob, NULL, CLSCTX_REMOTE_SERVER, &server, 1, &qi);
		if (FAILED(hr))
		{
			_tprintf(TEXT("Unable to connect to %s: "), host);
			PrintError(hr);
			return false;
		}
	}

	if (SUCCEEDED(qi.hr))
	{
		pJob = (IMPIJob*)qi.pItf;
		return true;
	}

	return false;
}

// Function name	: PrintOptions
// Description	    : 
// Return type		: void 
void PrintOptions()
{
	_tprintf(TEXT("Usage:\n"));
	_tprintf(TEXT("   msh host\n"));
}

// Function name	: GetAccountAndPassword
// Description	    : Attempts to read the password from the registry, 
//	                  upon failure it requests the user to provide one
// Return type		: void 
void GetAccountAndPassword()
{
	if (!ReadPasswordFromRegistry(g_pszAccount, g_pszPassword))
	{
		TCHAR ch=0;
		int index = 0;
		
		do
		{
			_ftprintf(stderr, TEXT("account: "));
			fflush(stderr);
			_getts(g_pszAccount);
		} while (_tcslen(g_pszAccount) == 0);
		
		_ftprintf(stderr, TEXT("password: "));
		fflush(stderr);

		HANDLE hStdin = GetStdHandle(STD_INPUT_HANDLE);
		DWORD dwMode;
		if (!GetConsoleMode(hStdin, &dwMode))
			dwMode = ENABLE_LINE_INPUT | ENABLE_PROCESSED_INPUT | ENABLE_MOUSE_INPUT;
		SetConsoleMode(hStdin, dwMode & ~ENABLE_ECHO_INPUT);
		_getts(g_pszPassword);
		SetConsoleMode(hStdin, dwMode);
		
		_ftprintf(stderr, TEXT("\n"));
	}
}

// Function name	: CtrlHandlerRoutine
// Description	    : 
// Return type		: BOOL WINAPI 
// Argument         : DWORD dwCtrlType
BOOL WINAPI CtrlHandlerRoutine(DWORD dwCtrlType)
{
	fprintf(stderr, "User break\n");
	if (g_pJob != NULL)
	{
		BSTR berror_msg = SysAllocString(L"");
		LONG error;
		//*
		fprintf(stderr, "Aborting job\n");
		g_pJob->Abort(&error, &berror_msg);
		SysFreeString(berror_msg);
		g_pJob->Release();
	}
	ExitProcess(1);
	return TRUE;
		/*/
		fprintf(stderr, "Forwarding break signal\n");
		g_pJob->SendBreak(&error, &berror_msg);
		SysFreeString(berror_msg);
	}
	return TRUE;
	//*/
}

struct RedirectInputThreadArg
{
	HANDLE hEvent;
	IStream **ppStream;
};

char g_pBuffer[1024];
HANDLE g_hBufferEvent1 = CreateEvent(NULL, TRUE, FALSE, NULL);
HANDLE g_hBufferEvent2 = CreateEvent(NULL, TRUE, FALSE, NULL);
DWORD g_num_read = 0;

// Function name	: ReadStdinThread
// Description	    : 
// Return type		: void 
void ReadStdinThread()
{
	HANDLE hStdin = GetStdHandle(STD_INPUT_HANDLE);
	while (true)
	{
		if (!ReadFile(hStdin, g_pBuffer, 1024, &g_num_read, NULL))
			return;
		if (g_num_read == 0)
			return;
		ResetEvent(g_hBufferEvent2);
		SetEvent(g_hBufferEvent1);
		WaitForSingleObject(g_hBufferEvent2, INFINITE);
	}
}

// Function name	: RedirectInputThread
// Description	    : 
// Return type		: void 
// Argument         : RedirectInputThreadArg *arg
void RedirectInputThread(RedirectInputThreadArg *arg)
{
	IMPIJob *pJob=NULL;
	HRESULT hr=S_OK;
	HANDLE hObject[2];
	long error=0;
	BSTR berror_msg;
	DWORD ret_val;

	hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);

	berror_msg = SysAllocString(L"");
	CoGetInterfaceAndReleaseStream (*arg->ppStream, IID_IMPIJob, (void**) &pJob);
	delete arg->ppStream;

	DWORD dwThreadID;
	HANDLE hRSIThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)ReadStdinThread, NULL, 0, &dwThreadID);

	hObject[0] = arg->hEvent;
	hObject[1] = g_hBufferEvent1;

	berror_msg = SysAllocString(L"");
	while (true)
	{
		ret_val = WaitForMultipleObjects(2, hObject, FALSE, INFINITE);
		if (ret_val == WAIT_OBJECT_0+1)
		{
			if (g_num_read > 0)
			{
				SAFEARRAYBOUND bound;
				VARIANT vInput;
				void *pBuf;

				VariantInit(&vInput);
				bound.lLbound = 0;
				bound.cElements = g_num_read;
				vInput.vt = VT_UI1 | VT_ARRAY;
				vInput.parray = SafeArrayCreate(VT_UI1, 1, &bound);

				SafeArrayAccessData(vInput.parray, &pBuf);
				memcpy(pBuf, g_pBuffer, g_num_read);
				SafeArrayUnaccessData(vInput.parray);

				error = 0;
				hr = pJob->PutJobInput(vInput, 0, &error, &berror_msg);
				if (FAILED(hr))
				{
					VariantClear(&vInput);
					printf("PutJobInput failed: %d\n", hr);
					PrintError(hr);
					break;
				}
				if (error)
				{
					VariantClear(&vInput);
					if (wcslen(berror_msg) < 1)
						wprintf(L"PutJobInput failed: %d %s\n", error, berror_msg);
					else
						wprintf(L"PutJobInput failed: %s\n", berror_msg);
					break;
				}
				VariantClear(&vInput);
			}
			ResetEvent(g_hBufferEvent1);
			SetEvent(g_hBufferEvent2);
		}
		else
		{
			//printf("g_hFinishedEvent signalled\n");
			break;
		}
	}

	pJob->Release();
	CoUninitialize();
}

// Function name	: main
// Description	    : 
// Return type		: void 
// Argument         : int argc
// Argument         : TCHAR *argv[]
void main(int argc, TCHAR *argv[])
{
	SetConsoleCtrlHandler(CtrlHandlerRoutine, TRUE);

	if (argc == 1)
	{
		PrintOptions();
		return;
	}

	_tcscpy(g_pszFirstHost, argv[1]);

	GetAccountAndPassword();

	if (argc == 2)
		_tcscpy(g_pszExe, TEXT("cmd.exe /Q"));
	else
	{
		_tcscpy(g_pszExe, TEXT("cmd.exe /c "));
		for (int i=2; i<argc; i++)
		{
			_tcscat(g_pszExe, argv[i]);
			if (i < argc-1)
				_tcscat(g_pszExe, TEXT(" "));
		}
	}

	TCHAR pszDir[MAX_PATH] = TEXT(".");
	GetCurrentDirectory(MAX_PATH, pszDir);

	LONG error = 0;
	BSTR berror_msg;
	BSTR bCommonArgs, bAccount, bPassword, bExe, bNic, bDir;
	HRESULT hr = S_OK;

	berror_msg = SysAllocString(L"");
#ifdef UNICODE
	bExe = SysAllocString(g_pszExe);
	bCommonArgs = SysAllocString(g_pszArgs);
	bAccount = SysAllocString(g_pszAccount);
	bPassword = SysAllocString(g_pszPassword);
	bNic = SysAllocString(g_pszFirstHost);
	bDir = SysAllocString(pszDir);
#else
	WCHAR wTemp[MAX_PATH];
	mbstowcs(wTemp, g_pszExe, strlen(g_pszExe)+1);
	bExe = SysAllocString(wTemp);
	mbstowcs(wTemp, g_pszAccount, strlen(g_pszAccount)+1);
	bAccount = SysAllocString(wTemp);
	mbstowcs(wTemp, g_pszPassword, strlen(g_pszPassword)+1);
	bPassword = SysAllocString(wTemp);
	mbstowcs(wTemp, g_pszArgs, strlen(g_pszArgs)+1);
	bCommonArgs = SysAllocString(wTemp);
	mbstowcs(wTemp, g_pszFirstHost, strlen(g_pszFirstHost)+1);
	bNic = SysAllocString(wTemp);
	mbstowcs(wTemp, pszDir, strlen(pszDir)+1);
	bDir = SysAllocString(wTemp);
#endif

	if (Connect(g_pszFirstHost, g_pJob))
	{
		error = 0;
		VariantInit(&g_vHosts);
		VariantInit(&g_vSMPInfo);
		hr = g_pJob->CreateJob(g_vHosts, g_vSMPInfo, 1, bNic, &error, &berror_msg);

		if (SUCCEEDED(hr))
		{
			if (error)
			{
				if (wcslen(berror_msg) < 1)
					wprintf(L"Unable to create the job: Error(%d), %s", error, berror_msg);
				else
					wprintf(L"Unable to create the job: %s", berror_msg);
				g_pJob->DestroyJob(&error, &berror_msg);
				g_pJob->Release();
				CoUninitialize();
				SysFreeString(berror_msg);
				SysFreeString(bExe);
				SysFreeString(bCommonArgs);
				SysFreeString(bAccount);
				SysFreeString(bPassword);
				SysFreeString(bNic);
				SysFreeString(bDir);
				return;
			}

			error = 0;
			hr = g_pJob->SetExecutable(bExe, &error, &berror_msg);
			if (FAILED(hr))
			{
				wprintf(L"SetExecutable failed: ");
				PrintError(hr);
				g_pJob->DestroyJob(&error, &berror_msg);
				g_pJob->Release();
				CoUninitialize();
				SysFreeString(berror_msg);
				SysFreeString(bExe);
				SysFreeString(bCommonArgs);
				SysFreeString(bAccount);
				SysFreeString(bPassword);
				SysFreeString(bNic);
				SysFreeString(bDir);
				return;
			}
			if (error)
			{
				if (wcslen(berror_msg) < 1)
					wprintf(L"SetExecutable failed: Error %d, %s\n", error, berror_msg);
				else
					wprintf(L"SetExecutable failed: %s\n", berror_msg);
				g_pJob->DestroyJob(&error, &berror_msg);
				g_pJob->Release();
				CoUninitialize();
				SysFreeString(berror_msg);
				SysFreeString(bExe);
				SysFreeString(bCommonArgs);
				SysFreeString(bAccount);
				SysFreeString(bPassword);
				SysFreeString(bNic);
				SysFreeString(bDir);
				return;
			}

			error = 0;
			hr = g_pJob->SetWorkingDirectory(bDir, &error, &berror_msg);
			if (FAILED(hr))
			{
				_tprintf(TEXT("DCOM failure: GetJobOutput()\n"));
				PrintError(hr);
				SysFreeString(berror_msg);
				SysFreeString(bExe);
				SysFreeString(bCommonArgs);
				SysFreeString(bAccount);
				SysFreeString(bPassword);
				SysFreeString(bNic);
				SysFreeString(bDir);
				return;
			}
			if (error)
			{
				if (wcslen(berror_msg) < 1)
					wprintf(L"SetWorkingDirectory failed: Error %d, %s\n", error, berror_msg);
				else
					wprintf(L"SetWorkingDirectory failed: %s\n", berror_msg);
				g_pJob->DestroyJob(&error, &berror_msg);
				g_pJob->Release();
				CoUninitialize();
				SysFreeString(berror_msg);
				SysFreeString(bExe);
				SysFreeString(bCommonArgs);
				SysFreeString(bAccount);
				SysFreeString(bPassword);
				SysFreeString(bNic);
				SysFreeString(bDir);
				return;
			}
			
			error = 0;
			hr = g_pJob->LaunchJob(bCommonArgs, bAccount, bPassword, MPI_LAUNCH_OPT_NOMPI, &error, &berror_msg);
			
			if (SUCCEEDED(hr))
			{
				if (error)
				{
					if (wcslen(berror_msg) < 1)
						wprintf(L"LaunchJob failed: Error %d, %s\n", error, berror_msg);
					else
						wprintf(L"LaunchJob failed: %s\n", berror_msg);
					g_pJob->DestroyJob(&error, &berror_msg);
					g_pJob->Release();
					CoUninitialize();
					SysFreeString(berror_msg);
					SysFreeString(bExe);
					SysFreeString(bCommonArgs);
					SysFreeString(bAccount);
					SysFreeString(bPassword);
					SysFreeString(bNic);
					SysFreeString(bDir);
					return;
				}

				RedirectInputThreadArg *pArg = new RedirectInputThreadArg;
				pArg->hEvent = g_hFinishedEvent;
				pArg->ppStream = new IStream*;
				// Marshall pJob to a thread which redirects user input
				hr = CoMarshalInterThreadInterfaceInStream(IID_IMPIJob, g_pJob, pArg->ppStream);
				if (FAILED(hr))
				{
					g_pJob->Abort(&error, &berror_msg);
					g_pJob->Release();
					SysFreeString(berror_msg);
					printf("CoMarshalInterThreadInterfaceInStream failed.\n");
					CoUninitialize();
					delete pArg;
					return;
				}

				DWORD dwThreadID;
				HANDLE hRedirectInputThread = CreateThread(
					NULL, 0, 
					(LPTHREAD_START_ROUTINE)RedirectInputThread,
					pArg, 
					0, &dwThreadID);
				
				if (hRedirectInputThread == NULL)
				{
					error = GetLastError();
					printf("CreateThread failed: error %d\n", error);
					g_pJob->DestroyJob(&error, &berror_msg);
					g_pJob->Release();
					SysFreeString(berror_msg);
					CoUninitialize();
					delete pArg;
					return;
				}

				HANDLE hStdout;
				hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
				if (hStdout == INVALID_HANDLE_VALUE)
				{
					error = GetLastError();
					_tprintf(TEXT("GetStdHandle failed: Error %d\n"), error);
					g_pJob->DestroyJob(&error, &berror_msg);
					g_pJob->Release();
					CoUninitialize();
					SysFreeString(berror_msg);
					SysFreeString(bExe);
					SysFreeString(bCommonArgs);
					SysFreeString(bAccount);
					SysFreeString(bPassword);
					SysFreeString(bNic);
					SysFreeString(bDir);
					return;
				}

				long more = 1, from = 0;
				VARIANT *v = new VARIANT;
				DWORD num_written, num_elements;
				void *pBuf;
				
				v->vt = VT_UI1 | VT_ARRAY;
				SAFEARRAYBOUND bound;
				bound.lLbound = 0;
				bound.cElements = 0;
				v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
				
				while (more == 1)
				{
					error = 0;
					hr = g_pJob->GetJobOutput(v, &from, &more, &error, &berror_msg);
					if (FAILED(hr))
					{
						_tprintf(TEXT("DCOM failure: GetJobOutput()\n"));
						PrintError(hr);
						SysFreeString(berror_msg);
						SysFreeString(bExe);
						SysFreeString(bCommonArgs);
						SysFreeString(bAccount);
						SysFreeString(bPassword);
						SysFreeString(bNic);
						SysFreeString(bDir);
						VariantClear(v);
						delete v;
						return;
					}
					if (error)
					{
						if (wcslen(berror_msg) < 1)
							wprintf(L"Error %d: %s", error, berror_msg);
						else
							wprintf(L"%s", berror_msg);
						SysFreeString(berror_msg);
						SysFreeString(bExe);
						SysFreeString(bCommonArgs);
						SysFreeString(bAccount);
						SysFreeString(bPassword);
						SysFreeString(bNic);
						SysFreeString(bDir);
						VariantClear(v);
						delete v;
						return;
					}
					if (v->parray != NULL)
					{
						num_elements = v->parray->rgsabound->cElements;
						if (num_elements > 0)
						{
							SafeArrayAccessData(v->parray, &pBuf);
							WriteFile(hStdout, pBuf, num_elements, &num_written, NULL);
							FlushFileBuffers(hStdout);
							SafeArrayUnaccessData(v->parray);
							// Destroy the array.
							SafeArrayDestroy(v->parray);
							// Create a new and empty array.
							v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
						}
						else
						{
							SafeArrayDestroy(v->parray);
							v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
						}
					}
					else
					{
						v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
					}
				}
				
				SafeArrayDestroy(v->parray);
				delete v;
				
				// Signal the redirect input thread to terminate
				SetEvent(g_hFinishedEvent);
				WaitForSingleObject(hRedirectInputThread, 1000);
				CloseHandle(hRedirectInputThread);
			}
			else
			{
				_tprintf(TEXT("LaunchJob failed: "));
				PrintError(hr);
				fflush(stdout);
			}
			
			g_pJob->DestroyJob(&error, &berror_msg);
			g_pJob->Release();
		}
		else
		{
			_tprintf(TEXT("Unable to create a job: "));
			PrintError(hr);
		}

		CoUninitialize();
	}
	else
	{
		_tprintf(TEXT("Unable to connect to '%s'\n"), g_pszFirstHost);
	}

	SysFreeString(berror_msg);
	SysFreeString(bExe);
	SysFreeString(bCommonArgs);
	SysFreeString(bAccount);
	SysFreeString(bPassword);
	SysFreeString(bNic);
	SysFreeString(bDir);
}
