// MessageQueue.cpp: implementation of the MessageQueue class.
//
//////////////////////////////////////////////////////////////////////

#include "nt_tcp_shm_message_queue.h"
#include <stdio.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

// Function name	: MessageQueue::MessageQueue
// Description	    : 
// Return type		: 
MessageQueue::MessageQueue()
{
	//m_pHeap = NULL;
	// Insert a node for 0 because I know it is the most common tag
	m_pHeap = new MessageQueue::InternalNode;
	m_pHeap->tag = 0;
	m_pHeap->list = NULL;
	m_pHeap->posted = NULL;
	m_pHeap->right = NULL;
	m_pHeap->left = NULL;

	m_num_elements = MSGQ_ELEMENT_BLOCK;
	m_num_available = MSGQ_ELEMENT_BLOCK;
	m_cur_index = 0;
	m_pool_size = 1;

	m_pPool = new MsgQueueElement*;
	m_pPool[0] = new MsgQueueElement[MSGQ_ELEMENT_BLOCK];

	for (int i=0; i<MSGQ_ELEMENT_BLOCK; i++)
	{
		m_pPool[0][i].hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
		m_pPool[0][i].bIn_use = false;
	}

	// Why is the following function protected in the header file by (_WIN32_WINNT >= 0x0500)
	// when the help files claim that they are implemented in NT 4?
	//InitializeCriticalSectionAndSpinCount(&m_CriticalSection, 4000);
	InitializeCriticalSection(&m_CriticalSection);
}

// Function name	: MessageQueue::DeleteHeapList
// Description	    : 
// Return type		: void 
// Argument         : MessageQueue::InternalListNode *listnode
void MessageQueue::DeleteHeapList(MessageQueue::InternalListNode *listnode)
{
	if (listnode == NULL)
		return;
	DeleteHeapList(listnode->next);
	delete listnode;
}

// Function name	: MessageQueue::DeleteHeap
// Description	    : 
// Return type		: void 
// Argument         : MessageQueue::InternalNode *node
void MessageQueue::DeleteHeap(MessageQueue::InternalNode *node)
{
	if (node == NULL)
		return;
	DeleteHeap(node->left);
	DeleteHeap(node->right);
	DeleteHeapList(node->list);
	DeleteHeapList(node->posted);
	delete node;
}

// Function name	: MessageQueue::~MessageQueue
// Description	    : 
// Return type		: 
MessageQueue::~MessageQueue()
{
	for (int i=0; i<m_num_elements; i++)
		CloseHandle(m_pPool[i/MSGQ_ELEMENT_BLOCK][i%MSGQ_ELEMENT_BLOCK].hEvent);
	m_num_elements = 0;
	m_num_available = 0;
	for (i=0; i<m_pool_size; i++)
		delete m_pPool[i];
	delete m_pPool;
	DeleteCriticalSection(&m_CriticalSection);
	DeleteHeap(m_pHeap);
	m_pHeap = NULL;
}

// Function name	: MessageQueue::allocElement
// Description	    : 
// Return type		: MessageQueue::MsgQueueElement* 
MessageQueue::MsgQueueElement* MessageQueue::allocElement()
{
	// Check to see if the allocated array needs to be expanded
	if (m_num_available == 0)
	{
		MessageQueue::MsgQueueElement **p = new MessageQueue::MsgQueueElement*[m_pool_size+1];
		for (int i=0; i<m_pool_size; i++)
			p[i] = m_pPool[i];
		p[m_pool_size] = new MessageQueue::MsgQueueElement[MSGQ_ELEMENT_BLOCK];
		for (i=0; i<MSGQ_ELEMENT_BLOCK; i++)
		{
			p[m_pool_size][i].bIn_use = false;
			p[m_pool_size][i].hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
		}
		m_pool_size++;
		delete m_pPool;
		m_pPool = p;
		m_num_elements += MSGQ_ELEMENT_BLOCK;
		m_num_available = MSGQ_ELEMENT_BLOCK;
	}

	// Find the next un-used element
	while (m_pPool[m_cur_index/MSGQ_ELEMENT_BLOCK][m_cur_index%MSGQ_ELEMENT_BLOCK].bIn_use)
		++m_cur_index %= m_num_elements;

	// Mark it as used
	m_pPool[m_cur_index/MSGQ_ELEMENT_BLOCK][m_cur_index%MSGQ_ELEMENT_BLOCK].bIn_use = true;

	m_num_available--;

	// Return it
	return &m_pPool[m_cur_index/MSGQ_ELEMENT_BLOCK][m_cur_index%MSGQ_ELEMENT_BLOCK];
}

// Function name	: MessageQueue::freeElement
// Description	    : 
// Return type		: void 
// Argument         : MessageQueue::MsgQueueElement *element
void MessageQueue::freeElement(MessageQueue::MsgQueueElement *element)
{
	EnterCriticalSection(&m_CriticalSection);
	element->bIn_use = false;
	ResetEvent(element->hEvent);
	m_num_available++;
	LeaveCriticalSection(&m_CriticalSection);
}

// Function name	: MessageQueue::FindNode
// Description	    : 
// Return type		: void 
// Argument         : MessageQueue::InternalNode *&node
void MessageQueue::FindNode(MessageQueue::InternalNode *&node)
{
	if (node == NULL)
	{
		node = new MessageQueue::InternalNode;
		node->list = new MessageQueue::InternalListNode;
		find_buffer = node->list->buffer = new char[find_length];
		node->list->length = find_length;
		node->list->from = find_from;
		find_pElement = node->list->element = allocElement();
		node->list->next = NULL;
		node->posted = NULL;
		node->left = NULL;
		node->right = NULL;
		node->tag = find_tag;
		return;
	}

	if (find_tag == node->tag)
	{
		if (node->posted != NULL)
		{
			//if (node->posted->length > 1024*1024) { printf("x");fflush(stdout); }
			find_buffer = node->posted->buffer;
			if (find_length > node->posted->length)
				node->posted->length = -1;
			else
				node->posted->length = find_length;
			node->posted->from = find_from;
			find_pElement = node->posted->element;
			node->posted = node->posted->next;
			return;
		}
		InternalListNode *list = node->list;
		if (list == NULL)
		{
			node->list = new InternalListNode;
			find_buffer = node->list->buffer = new char[find_length];
			node->list->length = find_length;
			node->list->from = find_from;
			find_pElement = node->list->element = allocElement();
			node->list->next = NULL;
			return;
		}
		while (list->next != NULL)
			list = list->next;
		list->next = new InternalListNode;
		list = list->next;
		find_buffer = list->buffer = new char[find_length];
		list->length = find_length;
		list->from = find_from;
		find_pElement = list->element = allocElement();
		list->next = NULL;
		return;
	}

	if (find_tag < node->tag)
		FindNode(node->left);
	else
		FindNode(node->right);
}

// Function name	: MessageQueue::GetBuffer
// Description	    : 
// Return type		: void* 
// Argument         : int tag
// Argument         : int length
// Argument         : int from
// Argument         : MessageQueue::MsgQueueElement **ppElement
void* MessageQueue::GetBuffer(int tag, int length, int from, MessageQueue::MsgQueueElement **ppElement)
{
	void *ret_val;
	EnterCriticalSection(&m_CriticalSection);
	find_tag = tag;
	find_length = length;
	find_from = from;
	FindNode(m_pHeap);
	*ppElement = find_pElement;
	ret_val = find_buffer;
	LeaveCriticalSection(&m_CriticalSection);
	return ret_val;
}

// Function name	: MessageQueue::FillFindNode
// Description	    : 
// Return type		: void 
// Argument         : MessageQueue::InternalNode *&node
void MessageQueue::FillFindNode(MessageQueue::InternalNode *&node)
{
	if (node == NULL)
	{
		//printf(".");fflush(stdout);
		node = new MessageQueue::InternalNode;
		node->posted = new MessageQueue::InternalListNode;
		node->posted->buffer = find_buffer;
		node->posted->length = find_length;
		node->posted->from = -1;
		find_pElement = node->posted->element = allocElement();
		node->posted->next = NULL;
		node->list = NULL;
		node->left = NULL;
		node->right = NULL;
		node->tag = find_tag;
		fillfind_node = node->posted;
		return;
	}

	if (find_tag == node->tag)
	{
		if (node->list != NULL)
		{
			find_length = node->list->length;
			fillfind_node = node->list;
			node->list = node->list->next;
			return;
		}
		//printf(".");fflush(stdout);
		InternalListNode *posted = node->posted;
		if (posted == NULL)
		{
			node->posted = new InternalListNode;
			node->posted->buffer = find_buffer;
			node->posted->length = find_length;
			node->posted->from = -1;
			find_pElement = node->posted->element = allocElement();
			fillfind_node = node->posted;
			node->posted->next = NULL;
			return;
		}
		while (posted->next != NULL)
			posted = posted->next;
		posted->next = new InternalListNode;
		posted = posted->next;
		posted->next = NULL;
		posted->buffer = find_buffer;
		posted->length = find_length;
		posted->from = -1;
		find_pElement = posted->element = allocElement();
		fillfind_node = posted;
		return;
	}

	if (find_tag < node->tag)
		FillFindNode(node->left);
	else
		FillFindNode(node->right);
}

// Function name	: MessageQueue::FillBuffer
// Description	    : 
// Return type		: bool 
// Argument         : int tag
// Argument         : void *buffer
// Argument         : int *length
// Argument         : int *from
bool MessageQueue::FillBuffer(int tag, void *buffer, int *length, int *from)
{
	bool bDeleteNeeded = true;
	InternalListNode *node;
	int ret_length;

	EnterCriticalSection(&m_CriticalSection);

	// Find the node which contains the data
	// If the node doesn't exist then an empty one is created
	find_tag = tag;
	find_buffer = buffer;
	find_length = *length;
	FillFindNode(m_pHeap);

	node = fillfind_node;
	ret_length = find_length;
	
	LeaveCriticalSection(&m_CriticalSection);

	// Wait for the buffer to be filled by another thread
	WaitForSingleObject(node->element->hEvent, INFINITE);

	// After the buffer has been filled then the from field is valid
	*from = node->from;
	ret_length = node->length;
	if (node->buffer == buffer)
	{
		if (ret_length == -1)
			return false;
		bDeleteNeeded = false;
	}
	else
	{
		if (ret_length > *length)
			return false;
		memcpy(buffer, node->buffer, ret_length);
	}
	*length = ret_length;

	// Return the element to the pool
	freeElement(node->element);

	// Delete the buffer if it was allocated
	if (bDeleteNeeded)
		delete node->buffer;
	delete node;

	return true;
}

// It was decided that the asyncronous handle would be an array of four integers (thus int *pID).
// I use the first as a pointer to an InternalListNode (only valid on 32 bit machines).
// The second is a void pointer to the user buffer.
// The third is the length - valid on both input and output.
// The fourth stores who the message was from.
bool MessageQueue::PostRecvBuffer(int tag, void *buffer, int length, int *pID)
{
	EnterCriticalSection(&m_CriticalSection);

	// Find the node which contains the data
	// If the node doesn't exist then an empty one is created
	find_tag = tag;
	find_buffer = buffer;
	find_length = length;
	FillFindNode(m_pHeap);

	pID[0] = (int)fillfind_node;
	pID[1] = (int)buffer;
	pID[2] = find_length >= length ? find_length : -1;
	pID[3] = -1;

	LeaveCriticalSection(&m_CriticalSection);

	return (pID[2] != -1);
}

bool MessageQueue::Wait(int *pID)
{
	//bool bDeleteNeeded = true;
	InternalListNode *node = (InternalListNode *)(pID[0]);

	if (node == NULL)
		return true;

	WaitForSingleObject(node->element->hEvent, INFINITE);

	// After the buffer has been filled then the from field is valid
	pID[3] = node->from;
	
	if (node->buffer == (void*)(pID[1]))
	{
		if (pID[2] == -1)
			return false;
		pID[2] = node->length;
		//bDeleteNeeded = false;
	}
	else
	{
		if (node->length > pID[2])
			return false;
		memcpy((void*)(pID[1]), node->buffer, node->length);
	}

	// Return the element to the pool
	freeElement(node->element);

	// Delete the buffer if it was allocated
	//if (bDeleteNeeded)
	if (node->buffer != (void*)(pID[1]))
		delete node->buffer;
	delete node;

	return true;
}

bool MessageQueue::Test(int *pID)
{
	if (pID[0])
	{
		if (WaitForSingleObject(((InternalListNode *)(pID[0]))->element->hEvent, 0) == WAIT_OBJECT_0)
		{
			//bool bDeleteNeeded = true;
			InternalListNode *node = (InternalListNode *)(pID[0]);
			
			// After the buffer has been filled then the from field is valid
			pID[3] = node->from;
			
			if (node->buffer == (void*)(pID[1]))
			{
				if (pID[2] == -1)
					return false;
				pID[2] = node->length;
				//bDeleteNeeded = false;
			}
			else
			{
				if (node->length > pID[2])
					return false;
				memcpy((void*)(pID[1]), node->buffer, node->length);
			}
			
			// Return the element to the pool
			freeElement(node->element);
			
			// Delete the buffer if it was allocated
			//if (bDeleteNeeded)
			if (node->buffer != (void*)(pID[1]))
				delete node->buffer;
			delete node;
			
			return true;
		}
		return false;
	}
	return true;
}

// Function name	: MessageQueue::FindAvailable
// Description	    : 
// Return type		: bool 
// Argument         : MessageQueue::InternalNode *node
// Argument         : int tag
// Argument         : int &from
bool MessageQueue::FindAvailable(MessageQueue::InternalNode *node, int tag, int &from)
{
	if (node == NULL)
		return false;

	if (tag == node->tag)
	{
		if (node->list == NULL)
			return false;
		from = node->list->from;
		return true;
	}

	if (tag < node->tag)
		return FindAvailable(node->left, tag, from);

	if (tag > node->tag)
		return FindAvailable(node->right, tag, from);

	return false;
}

// Function name	: MessageQueue::Available
// Description	    : 
// Return type		: bool 
// Argument         : int tag
// Argument         : int &from
bool MessageQueue::Available(int tag, int &from)
{
	bool ret_val;
	EnterCriticalSection(&m_CriticalSection);
	ret_val = FindAvailable(m_pHeap, tag, from);
	LeaveCriticalSection(&m_CriticalSection);
	return ret_val;
}
